from typing import Callable


def merge_dicts(dicts, precedence_dicts):
    """
    Merge multiple dictionaries with identical keys and structure into one dictionary,
    giving precedence to a list of dictionaries when a value is present or not None.

    Args:
        dicts (list): A list of dictionaries to merge.
        precedence_dicts (list): A list of dictionaries to give precedence to when a value is present or not None.

    Returns:
        dict: The merged dictionary.
    """
    if not dicts:
        return {}

    merged_dict = {}

    for key in dicts[0].keys():
        values = [d[key] for d in dicts if key in d]
        precedence_values = [d[key] for d in precedence_dicts if key in d and d[key] is not None]

        if isinstance(values[0], dict):
            # Recursive call for nested dictionaries
            merged_dict[key] = merge_dicts(
                [v for v in values if v is not None],
                [v for v in precedence_values if isinstance(v, dict)]
            )
        else:
            # Find the first non-None value from precedence dictionaries
            if precedence_values:
                merged_dict[key] = precedence_values[0]
            else:
                # Find the first non-None value from non-precedence dictionaries
                for value in values:
                    if value is not None:
                        merged_dict[key] = value
                        break

    return merged_dict


async def create_json_object(target: str, schema, get_value: Callable, current_obj=None, root_obj=None, parent_key='', prefix=''):
    if current_obj is None:
        current_obj = {}
        root_obj = {}

    for name, properties in schema.items():
        full_key = f'{parent_key}.{name}' if parent_key else name
        if properties['type'] == 'object':
            current_obj[name] = await create_json_object(target, properties['properties'], get_value, {}, root_obj, full_key)
        else:
            description = properties.get('description')
            enum = properties.get('enum', None)
            if prefix:
                description = prefix + (description[:1].lower() + description[1:])
            value = await get_value(target, full_key, properties['type'], description, enum, root_obj) if target else None
            current_obj[name] = value
            root_obj[full_key] = value

    return current_obj


def strip_non_numeric_chars(s):
    while len(s) > 0 and not s[0].isdigit():
        s = s[1:]
    while len(s) > 0 and not s[-1].isdigit():
        s = s[:-1]
    return s


from bpm_ai_core.ocr.ocr import OCR
from bpm_ai_core.speech_recognition.asr import ASRModel
from bpm_ai_core.util.file import is_supported_audio_file, is_file, is_supported_text_file
from bpm_ai_core.util.file import is_supported_img_file

from bpm_ai.common.errors import FileNotSupportedError


def prepare_images_for_llm_prompt(input_data: dict):
    """
    For multi-modal LLMs. Will be turned into Blob objects as part of the prompt processing
    and then into the respective image format of the LLM.
    """
    return {
        k: f"[# blob {v} #]"
        if (isinstance(v, str) and is_supported_img_file(v))
        else v for k, v in input_data.items()
    }


async def ocr_documents(input_data: dict, ocr: OCR | None = None):
    return {
        k: (await ocr.process(v)).full_text
        if (ocr and isinstance(v, str) and is_supported_img_file(v))
        else v for k, v in input_data.items()
    }


async def transcribe_audio(input_data: dict, asr: ASRModel | None = None) -> dict:
    return {
        k: (await asr.transcribe(v)).text
        if (asr and isinstance(v, str) and is_supported_audio_file(v))
        else v for k, v in input_data.items()
    }


def prepare_text_blobs(input_data: dict):
    """
    Will be turned into Blob objects as part of the prompt processing and then into text.
    """
    return {
        k: f"[# blob {v} #]"
        if (isinstance(v, str) and is_supported_text_file(v))
        else v for k, v in input_data.items()
    }


def assert_all_files_processed(input_data: dict):
    for v in input_data.values():
        if v and isinstance(v, str) and is_file(v):
            raise FileNotSupportedError(v)



from typing import Any

from bpm_ai_core.image_classification.image_classifier import ImageClassifier
from bpm_ai_core.llm.common.llm import LLM
from bpm_ai_core.ocr.ocr import OCR
from bpm_ai_core.prompt.prompt import Prompt
from bpm_ai_core.speech_recognition.asr import ASRModel
from bpm_ai_core.text_classification.text_classifier import TextClassifier
from bpm_ai_core.tracing.decorators import trace
from bpm_ai_core.util.file import is_supported_img_file
from bpm_ai_core.util.markdown import dict_to_md

from bpm_ai.common.errors import MissingParameterError
from bpm_ai.common.multimodal import transcribe_audio, prepare_images_for_llm_prompt, ocr_documents, prepare_text_blobs, \
    assert_all_files_processed
from bpm_ai.decide.schema import get_cot_decision_output_schema, get_decision_output_schema, remove_order_prefix_from_keys


@trace("bpm-ai-decide", ["llm"])
async def decide_llm(
    llm: LLM,
    input_data: dict[str, str | dict | None],
    instructions: str,
    output_type: str,
    possible_values: list[Any] | None = None,
    multiple_decision_values: bool = False,
    strategy: str | None = None,
    ocr: OCR | None = None,
    asr: ASRModel | None = None
) -> dict:
    if not instructions or instructions.isspace():
        raise MissingParameterError("question/instruction is required")
    if not output_type or output_type.isspace():
        raise MissingParameterError("output type is required")

    if all(value is None for value in input_data.values()):
        return {"decision": None, "reasoning": "No input values present."}

    if strategy == 'cot':
        output_schema = get_cot_decision_output_schema(output_type, possible_values, multiple_decision_values)
    else:
        output_schema = get_decision_output_schema(output_type, possible_values, multiple_decision_values)

    if not ocr and llm.supports_images():
        input_data = prepare_images_for_llm_prompt(input_data)
    else:
        input_data = await ocr_documents(input_data, ocr)
    input_data = await transcribe_audio(input_data, asr)
    input_data = prepare_text_blobs(input_data)
    assert_all_files_processed(input_data)

    prompt = Prompt.from_file(
        "decide",
        context=input_data,
        task=instructions,
        output_type=output_type,
        possible_values=possible_values,
        multiple_decision_values=multiple_decision_values,
        strategy=strategy
    )

    decide_schema = {
        "name": "store_decision",
        "description": f"Stores the final decision value{'s' if multiple_decision_values else ''} and corresponding reasoning.",
        "type": "object",
        "properties": output_schema
    }

    message = await llm.generate_message(prompt, output_schema=decide_schema)

    return remove_order_prefix_from_keys(message.content) if message.content else {}


@trace("bpm-ai-decide", ["classifier"])
async def decide_classifier(
    input_data: dict[str, str | dict | None],
    output_type: str,
    classifier: TextClassifier = None,
    image_classifier: ImageClassifier = None,
    question: str | None = None,
    possible_values: list[Any] | None = None,
    multiple_decision_values: bool = False,
    ocr: OCR | None = None,
    asr: ASRModel | None = None
) -> dict:
    if (classifier and image_classifier) or not (classifier or image_classifier):
        raise ValueError("Must provide either a TextClassifier or an ImageClassifier")
    if image_classifier and multiple_decision_values:
        raise ValueError("ImageClassifier does not support multiple decision values")

    zero_shot = (possible_values is not None) or output_type == "boolean"

    if not output_type or output_type.isspace():
        raise MissingParameterError("output type is required")
    if zero_shot and output_type == "boolean":
        possible_values = ["yes", "no"]
    if zero_shot:
        possible_values = [str(v) for v in possible_values]

    if all(value is None for value in input_data.values()):
        return {"decision": None, "reasoning": "No input values present."}

    if image_classifier and not classifier:
        image_path_or_url = list(input_data.values())[0]
        if len(input_data) > 1 or not isinstance(image_path_or_url, str) or not is_supported_img_file(image_path_or_url):
            raise ValueError("Must provide exactly one image file path variable as input when using an Image Classifier.")

        hypothesis_template = "Based on this image, the question '" + question + "' should be answered with '{}'" \
            if question else "This is an image of {}."

        classification = await image_classifier.classify(
            blob_or_path=image_path_or_url,
            classes=possible_values,
            hypothesis_template=hypothesis_template,
            confidence_threshold=0.1,
        )

    else:  # text classifier
        input_data = await ocr_documents(input_data, ocr)
        input_data = await transcribe_audio(input_data, asr)
        input_data = prepare_text_blobs(input_data)
        assert_all_files_processed(input_data)

        input_md = dict_to_md(input_data).strip()

        hypothesis_template = "In this example the question '" + question + "' should be answered with '{}'" \
            if question else "This example is {}."

        classification = await classifier.classify(
            input_md,
            possible_values,
            hypothesis_template=hypothesis_template,
            confidence_threshold=0.1,
            multi_label=multiple_decision_values
        )

    def raw_to_output_type(raw: str) -> Any:
        if output_type == "boolean":
            return (raw.lower() == 'yes' or raw.lower() == "true") if raw else None
        elif output_type == "integer":
            return int(raw) if raw else None
        elif output_type == "number":
            return float(raw) if raw else None
        else:
            return raw

    if not classification:
        result = None
    elif multiple_decision_values:
        result = [raw_to_output_type(label) for label, _ in classification.labels_scores]
    else:
        result = raw_to_output_type(classification.max_label)

    return {
        "decision": result,
        "reasoning": ""
    }



import re
from typing import Any

from bpm_ai_core.llm.common.blob import Blob
from bpm_ai_core.question_answering.question_answering import QuestionAnswering
from bpm_ai_core.llm.common.llm import LLM
from bpm_ai_core.ocr.ocr import OCR
from bpm_ai_core.prompt.prompt import Prompt
from bpm_ai_core.speech_recognition.asr import ASRModel
from bpm_ai_core.text_classification.text_classifier import TextClassifier
from bpm_ai_core.token_classification.zero_shot_token_classifier import ZeroShotTokenClassifier
from bpm_ai_core.tracing.decorators import trace
from bpm_ai_core.util.file import is_supported_img_file
from bpm_ai_core.util.json_schema import expand_simplified_json_schema
from bpm_ai_core.util.markdown import dict_to_md

from bpm_ai.common.errors import MissingParameterError
from bpm_ai.common.multimodal import transcribe_audio, prepare_images_for_llm_prompt, ocr_documents, prepare_text_blobs, \
    assert_all_files_processed
from bpm_ai.extract.util import merge_dicts, strip_non_numeric_chars, create_json_object


@trace("bpm-ai-extract", ["llm"])
async def extract_llm(
    llm: LLM,
    input_data: dict[str, str | dict | None],
    output_schema: dict[str, str | dict],
    multiple: bool = False,
    multiple_description: str = "",
    ocr: OCR | None = None,
    asr: ASRModel | None = None
) -> dict | list[dict]:
    if all(value is None for value in input_data.values()):
        return input_data

    if not ocr and llm.supports_images():
        input_data = prepare_images_for_llm_prompt(input_data)
    else:
        input_data = await ocr_documents(input_data, ocr)
    input_data = await transcribe_audio(input_data, asr)
    input_data = prepare_text_blobs(input_data)
    assert_all_files_processed(input_data)

    if not output_schema:
        return input_data

    def transform_result(extracted: dict):
        def empty_to_none(v):
            return None if v in ["", "null"] else v

        if multiple and "entities" in extracted.keys():
            extracted = extracted["entities"]

        if isinstance(extracted, list):
            return [transform_result(d) for d in extracted]
        else:
            return {k: empty_to_none(v) for k, v in extracted.items()}

    prompt = Prompt.from_file("extract", input=input_data)

    extract_schema = {
        "name": "information_extraction",
        "description": f"Extracts the relevant {'entities' if multiple else 'information'} from the passage.",
        "type": "object",
        "properties": output_schema if not multiple else {
            "entities": {"type": "array", "description": multiple_description, "items": output_schema}
        }
    }

    message = await llm.generate_message(prompt, output_schema=extract_schema)

    return transform_result(message.content or {})


@trace("bpm-ai-extract", ["extractive-qa"])
async def extract_qa(
    qa: QuestionAnswering,
    classifier: TextClassifier,
    input_data: dict[str, str | dict | None],
    output_schema: dict[str, str | dict],
    multiple: bool = False,
    multiple_description: str = "",
    ocr: OCR | None = None,
    vqa: QuestionAnswering | None = None,
    token_classifier: ZeroShotTokenClassifier | None = None,
    asr: ASRModel | None = None
) -> dict | list[dict]:
    if all(value is None for value in input_data.values()):
        return input_data

    if vqa:
        input_img_data = {k: v for k, v in input_data.items() if (isinstance(v, str) and is_supported_img_file(v))}
        input_data = {k: v for k, v in input_data.items() if k not in input_img_data.keys()}
    else:
        input_data = await ocr_documents(input_data, ocr)
    input_data = await transcribe_audio(input_data, asr)
    input_data = prepare_text_blobs(input_data)
    assert_all_files_processed(input_data)

    if not output_schema:
        return input_data

    input_md = dict_to_md(input_data).strip()
    output_schema = expand_simplified_json_schema(output_schema)["properties"]

    async def extract_value(text: str, field_name: str, field_type: str, description: str, enum: list, existing_values: dict) -> Any:
        """
        Extract value of type `field_type` from `text` based on `description`.
        `{}` placeholders in `description` will be formatted using `existing_values` dict which has flat dot notation keys
        (e.g. person.age if there is a person object with an age field).
        """
        if enum:
            # if an enum of values is given for the field, perform a classification instead of extraction
            return (await classifier.classify(text, enum)).max_label

        question = description + "?" if not description.endswith("?") else description
        question = question.format(**existing_values)
        question = question[:1].upper() + question[1:]  # capitalize first word

        if vqa and is_supported_img_file(text):
            model = vqa
            context = Blob.from_path_or_url(text)
        else:
            model = qa
            context = text

        qa_result = await model.answer(context, question, confidence_threshold=0.01 if not vqa else 0.1)

        if qa_result is None or qa_result.answer is None:
            return None

        if field_type == "integer":
            try:
                return int(strip_non_numeric_chars(qa_result.answer))
            except ValueError:
                return None
        elif field_type == "number":
            try:
                return float(strip_non_numeric_chars(qa_result.answer))
            except ValueError:
                return None
        else:
            return qa_result.answer.strip(" .,;:!?")

    if not multiple:
        result_dict = await create_json_object(input_md, output_schema, extract_value)
        if vqa:
            img_result_dicts = [
                await create_json_object(img, output_schema, extract_value) for img in input_img_data.values()
            ]
            # visual models can't process text and text models can't process documents, so if both modalities
            # are present we use crude merging of multiple result dicts, giving precedence to visual results
            return merge_dicts([result_dict], precedence_dicts=img_result_dicts)
        else:
            return result_dict
    else:
        if not multiple_description or multiple_description.isspace():
            raise MissingParameterError("Description for entity type is required.")

        result = await token_classifier.classify(input_md, classes=[multiple_description], confidence_threshold=0.75)
        entities = [s.word for s in result.spans]

        # to specify the current entity we are interested in, we mark it in the context and prepend a hint to the description
        description_prefix = f"For the {multiple_description} marked by << >>, "
        extracted = [
            await create_json_object(input_md.replace(entity, f"<< {entity} >>"), output_schema, extract_value, prefix=description_prefix)
            for entity in entities
        ]

        def clean_dict_strings(d):
            for key, value in d.items():
                if isinstance(value, str):
                    d[key] = re.sub(r'<<\s|\s>>', '', value)
            return d

        extracted = [clean_dict_strings(e) for e in extracted]

        # deduplicate
        extracted = [e for i, e in enumerate(extracted) if e not in extracted[:i]]
        # return only objects with at least one field having a value
        return [e for e in extracted if any(e.values())]




import re

from bpm_ai_core.util.linguistics import stopwords, replace_diacritics


def desc_to_var_name(desc: str):
    desc = desc.lower()
    v = remove_stop_words(desc, separator='_')
    v = replace_diacritics(v)
    return re.sub(r'[^a-zA-Z0-9_-]+', '', v)


def remove_stop_words(sentence, separator=' ', max_n_result_words: int = 6):
    # Split the sentence into individual words
    words = sentence.split()
    # Use a list comprehension to remove stop words
    filtered_words = [word for word in words if word not in stopwords]
    # Join the filtered words back into a sentence
    return separator.join(filtered_words[:max_n_result_words])


def type_to_prompt_type_str(type: str) -> str:
    match type:
        case "letter":
            return "emails and letters"
        case "chat":
            return "chat messages"
        case "social":
            return "social media posts"
        case "text":
            return "texts"
